#ifndef AVXASSEMBLY_H
#define AVXASSEMBLY_H

#include <MultiRegions/ContField.h>
#include <Collections/Collection.h>
#include <iostream>
#include "VecData.hpp"
#include "Operator.hpp"
#include "AVXUtil.hpp"

using namespace Nektar;

template<int VW>
struct AVXAssembly
{
public:
    AVXAssembly() = default;

    AVXAssembly(MultiRegions::AssemblyMapSharedPtr asmMap,
                MultiRegions::ExpListSharedPtr expList,
                const int nElmt)
        : m_asmMap    (asmMap),
          m_signChange(asmMap->GetSignChange()),
          m_nLocal    (asmMap->GetNumLocalCoeffs())
    {
        int nLocalPad = nElmt * expList->GetExp(0)->GetNcoeffs();

        // Copy & pad local to global array
        m_l2g = Array<OneD, int>(nLocalPad, asmMap->GetNumGlobalCoeffs());
        Vmath::Vcopy(asmMap->GetNumLocalCoeffs(),
                     asmMap->GetLocalToGlobalMap(), 1, m_l2g, 1);
        TransposeData(nElmt, VW, m_l2g);

        // Coyp & pad local to global sign array
        if(m_signChange)
        {
            m_l2gSign = Array<OneD, double>(nLocalPad, 0.0);
            Vmath::Vcopy(asmMap->GetNumLocalCoeffs(),
                         asmMap->GetLocalToGlobalSign(), 1, m_l2gSign, 1);
            TransposeData(nElmt, VW, m_l2gSign);
        }
    }

    template<int NM>
    inline int AssembleBlock(const double* local, double* global, int map_index)
    {
        if (m_signChange)
        {
            for (int i = 0; i < NM * VW; ++i)
            {
                global[m_l2g[map_index + i]] += m_l2gSign[map_index + i]
                    * local[i];
            }
            map_index += NM * VW;

        }
        else
        {
            for (int i = 0; i < NM * VW; ++i)
            {
                global[m_l2g[map_index + i]] += local[i];
            }
            map_index += NM * VW;
        }

        return map_index;
    }

    void Assemble(const Array<OneD, const NekDouble> &local,
                  Array<OneD, NekDouble> &global)
    {
        const int nLocalPad = m_l2g.size();

        if(m_signChange)
        {
            for (int i = 0; i < nLocalPad; ++i)
            {
                global[m_l2g[i]] += m_l2gSign[i] * local[i];
            }
        }
        else
        {
            for (int i = 0; i < nLocalPad; ++i)
            {
                global[m_l2g[i]] += local[i];
            }
        }
        m_asmMap->UniversalAssemble(global);
    }

    template<int NM>
    inline void ScatterBlock(const double* global,
                             double *local,
                             int map_index)
    {
        if(m_signChange)
        {
            for (int i = 0; i < NM * VW; ++i)
            {
                local[i] = m_l2gSign[map_index + i] *
                    global[m_l2g[map_index + i]];
            }
            map_index += NM * VW;
        }
        else
        {
            for (int i = 0; i < NM * VW; ++i)
            {
                local[i] = global[m_l2g[map_index + i]];
            }
            map_index += NM * VW;
        }
    }

    void Scatter(const Array<OneD, const NekDouble> &global,
                 Array<OneD, NekDouble> &local)
    {
        const int nLocalPad = m_l2g.size();

        if(m_signChange)
        {
            for (int i = 0; i < nLocalPad; ++i)
            {
                local[i] = m_l2gSign[i] * global[m_l2g[i]];
            }
        }
        else
        {
            for (int i = 0; i < nLocalPad; ++i)
            {
                local[i] = global[m_l2g[i]];
            }
        }
    }

    MultiRegions::AssemblyMapSharedPtr GetAssemblyMap()
    {
        return m_asmMap;
    }

    int GetNGlobal()
    {
        return m_asmMap->GetNumGlobalCoeffs();
    }

    MultiRegions::AssemblyMapSharedPtr m_asmMap;
    Array<OneD, int> m_l2g;
    Array<OneD, double> m_l2gSign;
    bool m_signChange;
    int m_nLocal;
};

#endif
