#include <cstdio>
#include <cstdlib>
#include <iomanip>

#include "Operator.hpp"
#include "AVXAssembly.h"
#include "AVXUtil.hpp"

#include <LibUtilities/Memory/NekMemoryManager.hpp>
#include <LibUtilities/BasicUtils/SessionReader.h>
#include <LibUtilities/Communication/Comm.h>
#include <LibUtilities/BasicUtils/Timer.h>
#include <MultiRegions/ContField.h>
#include <Collections/Collection.h>
#include <SpatialDomains/MeshGraph.h>

// Add likwid support
#ifdef LIKWID_PERFMON
#include <likwid.h>
#else
#define LIKWID_MARKER_INIT
#define LIKWID_MARKER_THREADINIT
#define LIKWID_MARKER_SWITCH
#define LIKWID_MARKER_REGISTER(regionTag)
#define LIKWID_MARKER_START(regionTag)
#define LIKWID_MARKER_STOP(regionTag)
#define LIKWID_MARKER_CLOSE
#define LIKWID_MARKER_GET(regionTag, nevents, events, time, count)
#endif

using namespace std;
using namespace Nektar;

bool solver_verbose = true;

template<int VW, class DataType>
class ContField
{
public:
    ContField(
        int                                  N,
        LibUtilities::SessionReaderSharedPtr session,
        SpatialDomains::MeshGraphSharedPtr   graph,
        bool                                 forceDeformed)
        : m_session(session), m_graph(graph)
    {
        // Accommodate use of orthogonal basis without having to specify in the
        // mesh file.
        graph->SetExpansionInfoToNumModes(N);

        MultiRegions::AssemblyMapSharedPtr asmMap;

        auto tmp = MemoryManager<MultiRegions::ContField>
            ::AllocateSharedPtr(session, graph);
        m_origAsmMap = tmp->GetLocalToGlobalMap();
        m_expList = tmp;

        m_nElmtOrig = m_expList->GetExpSize();
        m_nElmt     = m_nElmtOrig + 4 - (m_nElmtOrig % 4);
        m_nq        = m_expList->GetExp(0)->GetTotPoints() * m_nElmt;
        m_nm        = m_expList->GetExp(0)->GetNcoeffs() * m_nElmt;
        m_dim       = m_expList->GetExp(0)->GetShapeDimension();

        // Scan through list of elements and check if any are deformed. If they
        // are, we assume they _all_ have curvature for the purposes of testing.
        m_nqElmt = m_expList->GetExp(0)->GetTotPoints();
        m_nmElmt = m_expList->GetExp(0)->GetNcoeffs();

        auto shapeType = m_expList->GetExp(0)->DetShapeType();
        bool deformed = forceDeformed;
        bool deformed_fill = forceDeformed;

        for (int i = 0; i < m_nElmtOrig; ++i)
        {
            if (m_expList->GetExp(i)->GetMetricInfo()->GetGtype() ==
                SpatialDomains::eDeformed)
            {
                deformed = true;
                // For testing without an actual deformed mesh
                deformed_fill = false;
                break;
            }

            ASSERTL0(m_expList->GetExp(i)->DetShapeType() == shapeType,
                     "Shape type should be uniform within the mesh.");
            ASSERTL0(m_expList->GetExp(0)->GetTotPoints() == m_nqElmt,
                     "Number of quadrature points should be uniform within the mesh");
            ASSERTL0(m_expList->GetExp(0)->GetNcoeffs() == m_nmElmt,
                     "Number of modes should be uniform within the mesh");
        }

        // Assemble Jacobian list and derivative factors. Note this may be
        // downcast depending on DataType.
        if (deformed)
        {
            m_jac = Array<OneD, DataType>(m_nElmt * m_nqElmt, 0.0);
            m_df = Array<TwoD, DataType>(m_dim * m_dim, m_nElmt * m_nqElmt, 0.0);

            for (int i = 0; i < m_nElmtOrig; ++i)
            {
                auto jacElmt = m_expList->GetExp(i)->GetMetricInfo()->GetJac(
                    m_expList->GetExp(i)->GetPointsKeys());
                auto dfElmt = m_expList->GetExp(i)->GetMetricInfo()->GetDerivFactors(
                    m_expList->GetExp(i)->GetPointsKeys());

                for (int j = 0; j < m_nqElmt; ++j)
                {
                    int idx = deformed_fill ? 0 : j;
                    m_jac[i * m_nqElmt + j] = (DataType)jacElmt[idx];
                    for (int k = 0; k < m_dim * m_dim; ++k)
                    {
                        m_df[k][i * m_nqElmt + j] = dfElmt[k][idx];
                    }
                }
            }
        }
        else
        {
            m_jac = Array<OneD, NekDouble>(m_nElmt, 0.0);
            m_df = Array<TwoD, NekDouble>(m_dim * m_dim, m_nElmt, 0.0);

            for (int i = 0; i < m_nElmtOrig; ++i)
            {
                auto jacElmt = m_expList->GetExp(i)->GetMetricInfo()->GetJac(
                    m_expList->GetExp(i)->GetPointsKeys());
                auto dfElmt = m_expList->GetExp(i)->GetMetricInfo()->GetDerivFactors(
                    m_expList->GetExp(i)->GetPointsKeys());

                m_jac[i] = jacElmt[0];
                for(int j = 0; j < m_dim * m_dim; ++j)
                {
                    m_df[j][i] = dfElmt[j][0];
                }
            }
        }

        // Create AVX assembly map.
        m_asmMap = AVXAssembly<VW>(m_origAsmMap, m_expList, m_nElmt);

        // Basis vector.
        std::vector<LibUtilities::BasisSharedPtr> basis(m_dim);
        for(int i = 0; i < m_dim; i++)
        {
            basis[i] = m_expList->GetExp(0)->GetBasis(i);
        }

        // Generate operator string and create operator.
        m_helm = std::dynamic_pointer_cast<Helmholtz>(
            GetOperatorFactory().CreateInstance(
                CreateOpstring("Helmholtz", shapeType, deformed), basis, m_nElmt));
        m_helmGlobal = std::dynamic_pointer_cast<HelmholtzGlobal<VW>>(
            GetOperatorFactory().CreateInstance(
                CreateOpstring("HelmholtzGlobal", shapeType, deformed), basis, m_nElmt));

        m_helm->SetJac(m_jac);
        m_helm->SetDF(m_df);
        m_helmGlobal->SetJac(m_jac);
        m_helmGlobal->SetDF(m_df);
    }

    void BuildJacobiPreconditioner(Array<OneD, DataType> &jacobi)
    {
        // Conjugate gradient loop. Can use original unique map since this is
        // unchanged from AVX operations.
        auto &uniqueMap = m_origAsmMap->GetGlobalToUniversalMapUnique();

        const int nGlobal = m_origAsmMap->GetNumGlobalCoeffs();
        const int nBlocks = m_nElmt / VW;
        const int nm      = m_nElmt * m_nmElmt;

        // Firstly build a Jacobi preconditioner
        Array<OneD, NekDouble> tmpin(nm, 0.0), tmpout(nm), jacobi_local(nm);

        // Make sure Jacobi array is zeroed
        Vmath::Zero(nGlobal, &jacobi[0], 1);

        // Loop over number of modes in each block, and perform operator
        // evaluation for each dof so that we do not need to assemble matrices.
        for (int i = 0; i < m_nmElmt; ++i)
        {
            int offset = i * VW;
            for (int j = 0; j < nBlocks; ++j, offset += m_nmElmt * VW)
            {
                for (int k = 0; k < VW; ++k)
                {
                    tmpin[offset + k] = 1.0;
                }
            }

            if (m_session->GetComm()->GetRank() == 0)
            {
                std::cout << "precon step " << i << "/" << m_nmElmt << std::endl;
            }

            (*m_helm)(tmpin, tmpout);

            offset = i * VW;
            for (int j = 0; j < nBlocks; ++j, offset += m_nmElmt * VW)
            {
                for (int k = 0; k < VW; ++k)
                {
                    jacobi_local[offset+k] = tmpout[offset+k];
                    tmpin[offset+k] = 0.0;
                }
            }
        }

        // Assemble local contributions & calculate inverse
        m_asmMap.Assemble(jacobi_local, jacobi);

        for (int i = 0; i < nGlobal; ++i)
        {
            jacobi[i] = 1.0 / jacobi[i];
        }
    }

    MultiRegions::AssemblyMapSharedPtr &GetOrigAsmMap()
    {
        return m_origAsmMap;
    }

    const LibUtilities::CommSharedPtr &GetComm()
    {
        return m_session->GetComm();
    }

    std::shared_ptr<HelmholtzGlobal<VW>> &GetHelmGlobal()
    {
        return m_helmGlobal;
    }

    MultiRegions::ExpListSharedPtr &GetExpList()
    {
        return m_expList;
    }

    AVXAssembly<VW> &GetAsmMap()
    {
        return m_asmMap;
    }

protected:
    LibUtilities::SessionReaderSharedPtr m_session;
    SpatialDomains::MeshGraphSharedPtr   m_graph;
    MultiRegions::ExpListSharedPtr       m_expList;
    MultiRegions::AssemblyMapSharedPtr   m_origAsmMap;
    AVXAssembly<VW>                      m_asmMap;
    Array<OneD, DataType>                m_jac;
    Array<TwoD, DataType>                m_df;
    std::shared_ptr<Helmholtz>           m_helm;
    std::shared_ptr<HelmholtzGlobal<VW>> m_helmGlobal;

    /// Number of elements originally from expList
    int m_nElmtOrig;
    /// Number of elements, padded
    int m_nElmt;
    int m_nq;
    int m_nm;
    int m_dim;
    int m_nqElmt;
    int m_nmElmt;

    std::string CreateOpstring(
        std::string operatorType,
        LibUtilities::ShapeType shape,
        bool deformed = false)
    {
        std::string op_string = operatorType + "_";

        if(shape == LibUtilities::eTriangle)
        {
            op_string += "Tri";
        }
        else if(shape == LibUtilities::eQuadrilateral)
        {
            op_string += "Quad";
        }
        else if(shape == LibUtilities::eTetrahedron)
        {
            op_string += "Tet";
        }
        else if(shape == LibUtilities::ePrism)
        {
            op_string += "Prism";
        }
        else if(shape == LibUtilities::eHexahedron)
        {
            op_string += "Hex";
        }

        return op_string + (deformed ? "_Deformed" : "_Regular") + "_AVX";
    }

};

template<int VW, class DataType>
void ConjugateGradient(
    ContField<VW, DataType>            &field,
    const Array<OneD, const DataType>  &rhs,
    const Array<OneD, const DataType>  &jacobi,
    Array<OneD, DataType>              &sol,
    bool                                verbose = false)
{
    auto comm = field.GetComm();
    auto oper = field.GetHelmGlobal();
    const int nGlobal = field.GetOrigAsmMap()->GetNumGlobalCoeffs();
    auto &uniqueMap = field.GetOrigAsmMap()->GetGlobalToUniversalMapUnique();
    auto asmMap = field.GetAsmMap();

    // Zero initial solution field.
    Vmath::Zero(nGlobal, &sol[0], 1);

    Array<OneD, DataType> r(nGlobal), rnew(nGlobal);
    Array<OneD, DataType> z(nGlobal), znew(nGlobal), p(nGlobal), tmp(nGlobal);
    for (int i = 0; i < nGlobal; ++i)
    {
        r[i] = -rhs[i];
        z[i] = jacobi[i] * r[i];
        p[i] = z[i];
    }

    bool root = comm->GetRank() == 0;

    int k = 0;
    const int maxiter = 1000;
    LibUtilities::Timer t;
    t.Start();

    while (k < maxiter)
    {
        // Evaluate operator: tmp stores A*p
        (*oper)(p, tmp, asmMap);

        std::vector<DataType> alpha(2, 0.0), beta(2, 0.0);
        for (int i = 0; i < nGlobal; ++i)
        {
            if (uniqueMap[i] == 0)
            {
                continue;
            }

            alpha[0] += r[i] * z[i];
            alpha[1] += p[i] * tmp[i];
        }

        comm->AllReduce(alpha, LibUtilities::ReduceSum);
        DataType alph = alpha[0] / alpha[1];
        DataType res = 0.0;

        for (int i = 0; i < nGlobal; ++i)
        {
            sol[i] += alph * p[i];
            rnew[i] = r[i] - alph * tmp[i];
            if (uniqueMap[i] != 0)
            {
                res += rnew[i] * rnew[i];
            }
        }

        comm->AllReduce(res, LibUtilities::ReduceSum);
        if (verbose && comm->GetRank() == 0)
        {
            std::cout << "k = " << k << " res = " << res << std::endl;
        }

        if (res < 1e-10)
        {
            break;
        }

        for (int i = 0; i < nGlobal; ++i)
        {
            znew[i] = jacobi[i] * rnew[i];

            if (uniqueMap[i] == 0)
            {
                continue;
            }

            beta[0] += znew[i] * rnew[i];
            beta[1] += z[i] * r[i];
        }

        comm->AllReduce(beta, LibUtilities::ReduceSum);
        DataType bet = beta[0] / beta[1];

        for (int i = 0; i < nGlobal; ++i)
        {
            p[i] = znew[i] + bet * p[i];
            z[i] = znew[i];
            r[i] = rnew[i];
        }

        ++k;
    }
    t.Stop();

    // Calculate average elapsed time
    NekDouble elapsed = t.TimePerTest(k);
    comm->AllReduce(elapsed, LibUtilities::ReduceSum);
    elapsed /= comm->GetSize();

    NekDouble dof = oper->Ndof();
    comm->AllReduce(dof, LibUtilities::ReduceSum);

    if (comm->GetRank() == 0)
    {
        if (verbose)
        {
            std::cout << "time taken per iter (avg) = " << elapsed << std::endl;
            std::cout << "total dof = " << dof << std::endl;
            std::cout << "dof/s = " << dof / elapsed << std::endl;

            if (k == maxiter)
            {
                std::cout << "failed to converge" << std::endl;
            }
        }
        else
        {
            std::cout << "dof/s = " << dof / elapsed << std::endl;
        }
    }
}


int main(int argc, char *argv[])
{
    LibUtilities::SessionReader::RegisterCmdLineFlag(
        "data", "d", "Print in data format");
    LibUtilities::SessionReaderSharedPtr session
        = LibUtilities::SessionReader::CreateInstance(argc, argv);
    LibUtilities::CommSharedPtr comm = session->GetComm();

    LIKWID_MARKER_INIT;
    LIKWID_MARKER_THREADINIT;
    LIKWID_MARKER_REGISTER("Compute");

    //MultiRegions::VECTOR_WIDTH_GLOBAL = 4;

    // Read in mesh
    SpatialDomains::MeshGraphSharedPtr graph =
        SpatialDomains::MeshGraph::Read(session);

    int Ntest, order, verbose_p, deformed_p, ortho_p;
    session->LoadParameter("Ntest", Ntest, 1);
    session->LoadParameter("order", order, 3);
    session->LoadParameter("verbose", verbose_p, 1);
    session->LoadParameter("deformed", deformed_p, 0);
    session->LoadParameter("ortho", ortho_p, 0);

    solver_verbose = verbose_p != 0;
    bool deformed = deformed_p != 0;
    bool ortho = ortho_p != 0;

    bool fmt = session->DefinesCmdLineArgument("data");
    string sl = fmt ? "# " : "";

    // Create a field
    ContField<4, double> field(order, session, graph, deformed);

    ////////////////////// Set RHS forcing function
    auto expList = field.GetExpList();
    const int nq = expList->GetNpoints();
    const int nm = expList->GetNcoeffs();

    Array<OneD, NekDouble> xc(nq), yc(nq), zc(nq, 0.0);
    Array<OneD, NekDouble> tmp(nq), tmp2(nm), tmp3(nm), rhs(nm, 0.0);

    expList->GetCoords(xc, yc, zc);

    for (int i = 0; i < nq; ++i)
    {
        tmp[i] = sin(xc[i]) * cos(yc[i]) * cos(zc[i]);
    }

    // Inner product and assemble data for global RHS
    expList->IProductWRTBase(tmp, tmp2);
    field.GetOrigAsmMap()->Assemble(tmp2, rhs);

    const int nGlobal = field.GetOrigAsmMap()->GetNumGlobalCoeffs();
    Array<OneD, NekDouble> sol(nGlobal, 0.0), jacobi(nGlobal, 0.0);
    field.BuildJacobiPreconditioner(jacobi);

    ConjugateGradient(field, rhs, jacobi, sol, true);

    LIKWID_MARKER_CLOSE;
    comm->Finalise();

    return 0;
}
